'''
Red recurrente que trata de predecir la calificación de una reseña en
TripAdvisor analizando las palabras de la misma

En este caso se emplea la técnica del "word embedding" para codificar las
palabras de las reseñas
'''

import re
import numpy as np
import pandas as pd
from keras.models import Sequential
from keras.layers import Dense, Dropout, LSTM, Bidirectional, Embedding
from sklearn.feature_extraction.text import CountVectorizer

# Parámetros configurables
N = 20000   # Número de palabras del diccionario
T = 100     # Longitud prefijada de la reseña
D = 8       # Dimensión de los vectores de embedding
EPOCHS = 10
NEURONAS_CAPA = 50

# Función que construye un diccionario con los textos de entrada
def generar_diccionario(textos, sw, N):
    cv = CountVectorizer(stop_words = sw, max_features = N)
    cv.fit_transform(textos)
    diccionario = cv.vocabulary_
    diccionario = dict([(palabra, i+2) for i, palabra in enumerate(diccionario)])
    diccionario['DESC'] = 0
    diccionario['PAD'] = 1 
    return diccionario
    
# Función auxiliar para, dada una cadena, convertirla en una secuencia de 
# códigos según el diccionario aportado, hasta un tamaño máximo de T códigos
def procesar_cadena(texto, diccionario, stop_words, T):
    palabras = re.findall(r'\b\w+\b', texto.lower())
    palabras = list(filter(lambda x: x not in stop_words, palabras))
    resultado = [] 
    for i in range(0, T):
        if i < len(palabras): 
            if palabras[i] in diccionario:
                resultado.append(diccionario[palabras[i]])
            else:
                resultado.append(diccionario['DESC'])
        else:
            resultado.append(diccionario['PAD'])
    return np.array(resultado)

stop_words=['a', 'about', 'an', 'and', 'as', 'at', 'be', 'because', 'been', 
    'being', 'by', 'during', 'each', 'for', 'from', 'he', 'he\'d', 'he\'ll',
    'he\'s', 'her', 'him', 'his', 'how', 'i', 'i\'d', 'i\'ll', 'i\'m', 'i\'ve',
    'if', 'in', 'into', 'it', 'it\'s', 'its', 'let', 'let\'s', 'll', 'of', 'or', 
    'our', 'out', 'over', 're', 'same', 'she', 'she\'d', 'she\'ll', 'she\'s', 
    'so', 'some', 'such', 'than', 'that', 'that\'s', 'the', 'them', 'then', 
    'there', 'these', 'they', 'they\'d', 'they\'ll', 'they\'re', 'they\'ve', 
    'this', 'those', 'through', 'to', 'until', 've', 'very', 'we', 'we\'d', 
    'we\'ll', 'we\'re', 'we\'ve', 'what', 'what\'s', 'when', 'when\'s', 'where', 
    'where\'s', 'which', 'while', 'who', 'who\'s', 'whom', 'why', 'why\'s', 
    'with', 'would', 'wouldn', 'wouldn\'t', 'you', 'you\'d', 'you\'ll', 'you\'re', 
    'you\'ve', 'your', 'yours']

datos = pd.read_csv('tripadvisor_hotel_reviews.csv')

y = datos.iloc[:, -1]
num_clases = len(y.unique())
y = pd.get_dummies(datos['Rating'], columns=['Rating'])

textos = datos['Review']
X = []
diccionario = generar_diccionario(textos, stop_words, N)
for texto in textos:
    X.append(procesar_cadena(texto, diccionario, stop_words, T))
X = np.array(X)

# Construir la red
modelo = Sequential()
# Añadimos 2 unidades más al tamaño para incluir los códigos de
# palabras desconocidas y de padding
modelo.add(Embedding(input_dim=N+2, output_dim=D))
modelo.add(Bidirectional(LSTM(units=NEURONAS_CAPA, return_sequences=True)))
modelo.add(Bidirectional(LSTM(units=NEURONAS_CAPA)))
modelo.add(Dense(units=num_clases, activation='softmax'))

modelo.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
modelo.fit(X, y, validation_split=0.2, epochs=EPOCHS, batch_size=128)
